import os
import sys
import pygame

from . import engine
from .maths import *
from .data import read_json, load

TILESETS = {}

class Tile:
	def __init__(self, image, tiletype):
		self.image = image
		self.type = tiletype

class Tileset:
	def __init__(self, tiles):
		self.tiles = tiles

class Layer:
	def __init__(self, tiles, offset, background=False):
		self.tiles = tiles
		self.offset = offset
		self.background = background

class Tiledmap:
	def __init__(self, width, height, 
					   tileset, layers, layerids, 
						objects, lighting, stepprefix):
		self.width = width
		self.height = height
		self.tileset = tileset
		self.layers = layers
		self.layerids = layerids
		self.objects = objects
		self.surface = None
		self.back_surface = None
		self.lighted = bool(lighting)
		self.lighting = lighting or (0,0,0)
		self.stepprefix = stepprefix

	def get_tilecol(self, x, y):
		y = self.height-y-1
		if x < 0 or x >= self.width or y < 0 or y >= self.height:
			return None
		tileid = self.layers[1].tiles[y*self.width+x]
		if tileid == 0:
			return None
		return self.tileset.tiles[tileid].type != 'background'

	def set(self, x,y, layername, tileid, realset=True):
		unit = engine.UNIT
		layer = self.layers[self.layerids[layername]]
		tile = self.tileset.tiles[tileid]
		if realset:
			layer.tiles[y*self.width+x] = tileid
		self.surface.blit(tile.image, (int(x*unit), int(y*unit)))

	def get_tile(self, layer, x, y):
		y = self.height-y-1
		if x < 0 or x >= self.width or y < 0 or y >= self.height:
			return None
		tileid = self.layers[layer].tiles[y*self.width+x]
		if tileid == 0:
			return None
		return self.tileset.tiles[tileid]

	def build_image(self):
		unit = engine.UNIT
		#unit = 1
		tileset = self.tileset
		layers = self.layers
		width = self.width
		height = self.height

		surface = pygame.Surface((width*unit, height*unit), pygame.SRCALPHA)
		back_surface = pygame.Surface((width*unit, height*unit), pygame.SRCALPHA)

		for layer in layers:
			tiles = layer.tiles			
			offx, offy = layer.offset
			for y in range(height):
				for x in range(width):
					tileid = tiles[y*width+x]
					if tileid == 0:
						continue
					tile = tileset.tiles[tileid]
					if layer.background:
						back_surface.blit(tile.image, (int((x+offx)*unit), int((y+offy)*unit)))
					else:
						surface.blit(tile.image, (int((x+offx)*unit), int((y+offy)*unit)))

		self.surface = surface
		self.back_surface = back_surface

def load_tileset(filename):
	jsondata = read_json(filename)
	tilesdata = jsondata['tiles']

	tiles = {}
	for tileid in tilesdata:
		tiledata = tilesdata[tileid]
		imagefile = tiledata['image']
		image = pygame.image.load(load(imagefile), imagefile)
		image = pygame.transform.scale(image, (engine.UNIT, engine.UNIT)).convert_alpha()
		tiletype = tiledata.get('type', 'none')
		tile = Tile(image, tiletype)
		tiles[int(tileid)+1] = tile

	tileset = Tileset(tiles)
	TILESETS[filename] = tileset
	return tileset

def load_map(filename):
	UNIT = engine.UNIT
	TEXUNIT = engine.TEXUNIT
	jsondata = read_json(filename)
	width = jsondata['width']
	height = jsondata['height']
	tileset = jsondata['tilesets'][0]
	tileset_source = tileset['source']
	tileset_source = os.path.join(*os.path.split(filename)[:-1], tileset_source)

	if tileset_source in TILESETS:
		tileset = TILESETS[tileset_source]
	else:
		tileset = load_tileset(tileset_source)

	layersdata = jsondata['layers']
	layers = []
	layerids = {}
	objects = {'doors':[], 'drawers':[], 'cabinets':[], 'spawns':{}}
	lighting = None
	stepprefix = ''
	if 'properties' in jsondata:
		properties = jsondata['properties']
		if 'lighting' in properties:
			h = properties['lighting'].lstrip('#')
			lighting = tuple(int(h[i:i+2], 16) for i in (0, 2, 4))
		stepprefix = properties.get('stepprefix', stepprefix)
	for layerdata in layersdata:
		layertype = layerdata['type']
		name = layerdata['name']
		if layertype == 'tilelayer':
			data = layerdata['data']
			offx = 0.0
			offy = 0.0
			props = layerdata.get('properties')
			background = False
			if props:
				offx = props.get('offset_x', 0.0)
				offy = props.get('offset_y', 0.0)
				background = props.get('background', background)
			if name == 'furniture':
				for y in range(height):
					for x in range(width):
						tileid = data[y*width+x]
						if tileid == 0:
							continue
						tile = tileset.tiles[tileid]
						if tile.type == 'drawer':
							objects['drawers'].append((x,))
						if tile.type == 'cabinet':
							objects['cabinets'].append((x,y))
			layerids[name] = len(layers)
			layers.append(Layer(data, (offx, offy), background))
		elif layertype == 'objectgroup':
			layerobjects = layerdata['objects']
			for objdef in layerobjects:
				x = objdef['x']/TEXUNIT
				y = height-objdef['y']/TEXUNIT
				w = objdef['width']/TEXUNIT
				h = objdef['height']/TEXUNIT
				y -= h
				objtype = objdef['type']
				props = objdef['properties']
				if objtype == 'door':
					mapname = props['map']
					spawn = props['spawn']
					objects['doors'].append((x,y,w,h, mapname, spawn))
				elif objtype == 'spawn':
					objects['spawns'][props['id']] = (x,y)
		else:
			print("unknown layer type: %s" % layertype, file=sys.stderr)
		
	return Tiledmap(width, height, tileset, layers, layerids, objects, lighting, stepprefix)
